// 燃料等级：
// 1 - 最小单位，如： void，各种循环里插入的计费点
// 3 - 最基本的，如：一元运算符
// 5 - 二元运算符
// 10 - 比二元运算符复杂点的，如：纯函数调用
// 20 - 内置工具函数，包含纯函数调用+内部一些逻辑
// 50 - 知道消耗多，但不清楚具体多少倍，所以设置大一点，同时也是不推荐使用。如：new
// 自定义的

module.exports = {
  Contract: {
    register: 1000,
    transfer: 1000,
    pay: 500,
    storePerChar: 2
  },
  Operator: {
    UnaryExpression: {
      '+': 3,
      '-': 3,
      '!': 3,
      '~': 3,
      'typeof': 3,
      'void': 1
    },
    UpdateExpression: {
      '++': 5,
      '--': 5,
    },
    BinaryExpression: {
      '==': 5,
      '!=': 5,
      '===': 5,
      '!==': 5,
      '<': 5,
      '<=': 5,
      '>': 5,
      '>=': 5,
      '<<': 5,
      '>>': 5,
      '>>>': 5,
      '+': 5,
      '-': 5,
      '*': 5,
      '/': 5,
      '%': 5,
      '|': 5,
      '^': 5,
      '&': 5,
      'in': 5,
      'instanceof': 5,
    },
    AssignmentExpression: {
      '=': 3,
      '+=': 5,
      '-=': 5,
      '*=': 5,
      '/=': 5,
      '%=': 5,
      '<<=': 5,
      '>>=': 5,
      '>>>=': 5,
      '|=': 5,
      '^=': 5,
      '&=': 5,
    },
    LogicalExpression: {
      '&&': 5,
      '||': 5,
    },
    ConditionalExpression: 5,
    NewExpression: 50,
    // function fn (){}
    FunctionDeclaration: 10,
    // const fn = function() {}
    FunctionExpression: 10,
    ArrowFunctionExpression: 10,

    WhileStatement: 1,
    ForStatement: 1,
    ForInStatement: 1,
    ForOfStatement: 1,
    DoWhileStatement: 1,

    MemberExpression: 1,

    // RestElement: 50,
    SpreadElement: 50,

    VariableDeclaration: 3
  },
  assert: 20,
  Mapping: {
    fromItems: function (args) { return (args[0] && args[0].length) * 10 || 10 },
    isMapping: 20,
    keysOf: function (args) {
      return (args[0] && args[0][Symbol.for('map')] && args[0][Symbol.for('map')].size) * 10 || 10
    },
    // valuesOf: function (args) {
    //   return (args[0] && args[0][Symbol.for('map')] && args[0][Symbol.for('map')].size) * 10 || 10
    // },
    sizeOf: 20
  },
  Vector: {
    isVector: 20,
    push: function (args) { return args.length * 10 || 10 },
    pop: 20,
    size: 20
  },
  Crypto: {
    sha256: {
      hash: function (args) {
        const price = (args[0].byteLength || args[0].length) * 10
        return price >= 100 ? price : 100
      }
    },
    ripemd160: {
      hash: function (args) {
        const price = (args[0].byteLength || args[0].length) * 10
        return price >= 100 ? price : 100
      }
    },
    ed25519: {
      sign: function (args) { return args[0].byteLength * 10 }, // args[0] 一定大于10
      verify: function (args) { return args[0].byteLength * 10 }
    }
  },
  Util: {
    Address: {
      generateNormalAddress: 20
    }
  },
  ByteBuffer: {
    static: {
      wrap: function (args) {
        if (typeof args[0] === 'string') {
          return args[0].length * 10 || 10
        }
        return 20
      },
      fromBase64: function (args) {
        return args[0].length * 10 || 10
      },
      fromBinary: function (args) {
        return args[0].length * 10 || 10
      },
      fromHex: function (args) {
        return args[0].length * 10 || 10
      },
      fromUTF8: function (args) {
        return args[0].length * 10 || 10
      },
      isByteBuffer: 20
    },
    prototype: {
      readInt8: 20,
      readUint8: 20,
      readInt16: 20,
      readUint16: 20,
      readInt32: 20,
      readUint32: 20,
      readInt64: 20,
      readUint64: 20,
      readFloat32: 20,
      readFloat64: 20,
      readString: 20,
      readCString: 20,
      readIString: 20,
      writeInt8: 20,
      writeUint8: 20,
      writeInt16: 20,
      writeUint16: 20,
      writeInt32: 20,
      writeUint32: 20,
      writeInt64: 20,
      writeUint64: 20,
      writeFloat32: 20,
      writeFloat64: 20,
      writeString: function (args) { return args[0].length * 10 || 10 },
      writeCString: function (args) { return args[0].length * 10 || 10 },
      writeIString: function (args) { return args[0].length * 10 || 10 },
      append: 20,
      clone: 100,
      flip: 20,
      resize: 0, // 内部使用
      toBuffer: 20,
      toString: 50,
      toBase64: 50,
      toBinary: 50,
      toHex: 50,
      toUTF8: 50
    }
  },
  Object: {
    static: {
      assign: {
        before: function (args) { return args.length * 10 || 10 }
      },
      entries: {
        after: function (args, result) { return result.length * 10 || 10 }
      },
      keys: {
        after: function (args, result) { return result.length * 10 || 10 }
      },
      values: {
        after: function (args, result) { return result.length * 10 || 10 }
      },
      is: 20
    }
  },
  Array: {
    static: {
      from: {
        after: function (args, result) { return result.length * 10 || 10 }
      },
      isArray: 20,
      of: {
        after: function (args, result) { return result.length * 10 || 10 }
      }
    },
    prototype: {
      concat: {
        after: function (args, result) { return result.length * 10 || 10 }
      },
      copyWithin: {
        after: function (args, result) { return result.length * 10 || 10 }
      },
      every: {
        // 回调函数内本身有计费
        before: function () { return this.length ? 0 : 10 }
      },
      fill: {
        before: function () { return this.length * 10 || 10 }
      },
      filter: {
        // 回调函数内本身有计费
        before: function () { return this.length ? 0 : 10 }
      },
      find: {
        before: function () { return this.length * 10 || 10 }
      },
      findIndex: {
        before: function () { return this.length * 10 || 10 }
      },
      forEach: {
        // 回调函数内本身有计费
        before: function () { return this.length ? 0 : 10 }
      },
      includes: {
        before: function () { return this.length * 10 || 10 }
      },
      indexOf: {
        before: function () { return this.length * 10 || 10 }
      },
      join: {
        before: function () { return this.length * 10 || 10 }
      },
      lastIndexOf: {
        before: function () { return this.length * 10 || 10 }
      },
      map: {
        // 回调函数内本身有计费
        before: function () { return this.length ? 0 : 10 }
      },
      pop: 20,
      push: 20,
      reduce: {
        // 回调函数内本身有计费
        before: function () { return this.length ? 0 : 10 }
      },
      reduceRight: {
        // 回调函数内本身有计费
        before: function () { return this.length ? 0 : 10 }
      },
      reverse: {
        before: function () { return this.length * 10 || 10 }
      },
      shift: 20,
      unshift: 20,
      slice: {
        before: function () { return this.length * 10 || 10 }
      },
      some: {
        // 回调函数内本身有计费
        before: function () { return this.length ? 0 : 10 }
      },
      sort: {
        // 回调函数内本身有计费
        before: function () { return this.length ? 0 : 10 }
      },
      splice: {
        before: function () { return this.length * 10 || 10 }
      },
      toString: {
        before: function () { return this.length * 10 || 10 }
      }
    }
  },
  String: {
    static: {
      fromCharCode: 20,
    },
    prototype: {
      charAt: 20,
      charCodeAt: 20,
      startsWith: 20,
      endsWith: 20,
      includes: {
        before: function () { return this.length * 10 || 10 }
      },
      indexOf: {
        before: function () { return this.length * 10 || 10 }
      },
      lastIndexOf: {
        before: function () { return this.length * 10 || 10 }
      },
      padStart: {
        before: function (args) { return args[0] * 10 || 10 }
      },
      padEnd: {
        before: function (args) { return args[0] * 10 || 10 }
      },
      repeat: {
        after: function (args, result) { return result.length * 10 || 10 }
      },
      slice: {
        after: function (args, result) { return result.length * 10 || 10 }
      },
      substr: {
        after: function (args, result) { return result.length * 10 || 10 }
      },
      substring: {
        after: function (args, result) { return result.length * 10 || 10 }
      },
      split: {
        after: function (args, result) { return result.length * 10 || 10 }
      },
      toLowerCase: {
        before: function () { return this.length * 10 || 10 }
      },
      toUpperCase: {
        before: function () { return this.length * 10 || 10 }
      },
      trim: 20,
      trimLeft: 20,
      trimRight: 20,
      toString: 0
    }
  },
  Number: {
    static: {
      isNaN: 20,
      parseFloat: 20,
      parseInt: 20
    },
    prototype: {
      toExponential: 20,
      toFixed: 20,
      toPrecision: 20,
      toString: 20
    }
  },
  Math: {
    static: {
      abs: 20,
      acos: 20,
      acosh: 20,
      asin: 20,
      asinh: 20,
      atan: 20,
      atan2: 20,
      atanh: 20,
      cbrt: 20,
      ceil: 20,
      clz32: 20,
      cos: 20,
      cosh: 20,
      exp: 20,
      expm1: 20,
      floor: 20,
      fround: 20,
      hypot: 20,
      imul: 20,
      log: 20,
      log10: 20,
      log1p: 20,
      log2: 20,
      max: 20,
      min: 20,
      pow: 20,
      round: 20,
      sign: 20,
      sin: 20,
      sinh: 20,
      sqrt: 20,
      tan: 20,
      tanh: 20,
      trunc: 20
    }
  },
  BigInt: {
    prototype: {
      toString: 20
    }
  }
}
